@@ -46,7 +46,6 @@ const char dhcpcd_copyright[] = "Copyright (c) 2006-2015 Roy Marples";
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <syslog.h>
 #include <unistd.h>
 #include <time.h>
 
@@ -68,12 +67,13 @@ const char dhcpcd_copyright[] = "Copyright (c) 2006-2015 Roy Marples";
 
 #ifdef USE_SIGNALS
 const int dhcpcd_handlesigs[] = {
+	SIGTERM,
+	SIGINT,
 	SIGALRM,
 	SIGHUP,
-	SIGINT,
-	SIGPIPE,
-	SIGTERM,
 	SIGUSR1,
+	SIGUSR2,
+	SIGPIPE,
 	0
 };
 
@@ -192,7 +192,7 @@ handle_exit_timeout(void *arg)
 	struct dhcpcd_ctx *ctx;
 
 	ctx = arg;
-	syslog(LOG_ERR, "timed out");
+	logger(ctx, LOG_ERR, "timed out");
 	if (!(ctx->options & DHCPCD_MASTER)) {
 		eloop_exit(ctx->eloop, EXIT_FAILURE);
 		return;
@@ -261,20 +261,26 @@ dhcpcd_daemonise(struct dhcpcd_ctx *ctx)
 		return 0;
 	/* Setup a signal pipe so parent knows when to exit. */
 	if (pipe(sidpipe) == -1) {
-		syslog(LOG_ERR, "pipe: %m");
+		logger(ctx, LOG_ERR, "pipe: %m");
 		return 0;
 	}
-	syslog(LOG_DEBUG, "forking to background");
+	logger(ctx, LOG_DEBUG, "forking to background");
 	switch (pid = fork()) {
 	case -1:
-		syslog(LOG_ERR, "fork: %m");
+		logger(ctx, LOG_ERR, "fork: %m");
 		return 0;
 	case 0:
 		setsid();
+		/* Some polling methods don't survive after forking,
+		 * so ensure we can requeue all our events. */
+		if (eloop_requeue(ctx->eloop) == -1) {
+			logger(ctx, LOG_ERR, "eloop_requeue: %m");
+			eloop_exit(ctx->eloop, EXIT_FAILURE);
+		}
 		/* Notify parent it's safe to exit as we've detached. */
 		close(sidpipe[0]);
 		if (write(sidpipe[1], &buf, 1) == -1)
-			syslog(LOG_ERR, "failed to notify parent: %m");
+			logger(ctx, LOG_ERR, "failed to notify parent: %m");
 		close(sidpipe[1]);
 		if ((fd = open(_PATH_DEVNULL, O_RDWR, 0)) != -1) {
 			dup2(fd, STDIN_FILENO);
@@ -287,13 +293,13 @@ dhcpcd_daemonise(struct dhcpcd_ctx *ctx)
 		/* Wait for child to detach */
 		close(sidpipe[1]);
 		if (read(sidpipe[0], &buf, 1) == -1)
-			syslog(LOG_ERR, "failed to read child: %m");
+			logger(ctx, LOG_ERR, "failed to read child: %m");
 		close(sidpipe[0]);
 		break;
 	}
 	/* Done with the fd now */
 	if (pid != 0) {
-		syslog(LOG_INFO, "forked to background, child pid %d", pid);
+		logger(ctx, LOG_INFO, "forked to background, child pid %d", pid);
 		write_pid(ctx->pid_fd, pid);
 		close(ctx->pid_fd);
 		ctx->pid_fd = -1;
@@ -312,7 +318,7 @@ stop_interface(struct interface *ifp)
 	struct dhcpcd_ctx *ctx;
 
 	ctx = ifp->ctx;
-	syslog(LOG_INFO, "%s: removing interface", ifp->name);
+	logger(ctx, LOG_INFO, "%s: removing interface", ifp->name);
 	ifp->options->options |= DHCPCD_STOPPING;
 
 	dhcp6_drop(ifp, NULL);
@@ -375,6 +381,10 @@ configure_interface1(struct interface *ifp)
 	    !(ifp->ctx->options & DHCPCD_TEST))
 		ifo->options |= DHCPCD_IPV6RA_OWN;
 
+	/* If we're a psuedo interface, ensure we disable as much as we can */
+	if (ifp->options->options & DHCPCD_PFXDLGONLY)
+		ifp->options->options &= ~(DHCPCD_IPV4 | DHCPCD_IPV6RS);
+
 	/* We want to disable kernel interface RA as early as possible. */
 	if (ifo->options & DHCPCD_IPV6RS) {
 		/* If not doing any DHCP, disable the RDNSS requirement. */
@@ -468,10 +478,12 @@ configure_interface1(struct interface *ifp)
 	}
 
 #ifdef INET6
-	if (ifo->ia_len == 0 && ifo->options & DHCPCD_IPV6) {
+	if (ifo->ia_len == 0 && ifo->options & DHCPCD_IPV6 &&
+	    ifp->name[0] != '\0')
+	{
 		ifo->ia = malloc(sizeof(*ifo->ia));
 		if (ifo->ia == NULL)
-			syslog(LOG_ERR, "%s: %m", __func__);
+			logger(ifp->ctx, LOG_ERR, "%s: %m", __func__);
 		else {
 			ifo->ia_len = 1;
 			ifo->ia->ia_type = D6_OPTION_IA_NA;
@@ -508,19 +520,21 @@ dhcpcd_selectprofile(struct interface *ifp, const char *profile)
 		r = print_string(pssid, sizeof(pssid), ESCSTRING,
 		    ifp->ssid, ifp->ssid_len);
 		if (r == -1) {
-			syslog(LOG_ERR, "%s: %s: %m", ifp->name, __func__);
+			logger(ifp->ctx, LOG_ERR,
+			    "%s: %s: %m", ifp->name, __func__);
 			pssid[0] = '\0';
 		}
 	} else
 		pssid[0] = '\0';
 	ifo = read_config(ifp->ctx, ifp->name, pssid, profile);
 	if (ifo == NULL) {
-		syslog(LOG_DEBUG, "%s: no profile %s", ifp->name, profile);
+		logger(ifp->ctx, LOG_DEBUG, "%s: no profile %s",
+		    ifp->name, profile);
 		return -1;
 	}
 	if (profile != NULL) {
 		strlcpy(ifp->profile, profile, sizeof(ifp->profile));
-		syslog(LOG_INFO, "%s: selected profile %s",
+		logger(ifp->ctx, LOG_INFO, "%s: selected profile %s",
 		    ifp->name, profile);
 	} else
 		*ifp->profile = '\0';
@@ -532,11 +546,13 @@ dhcpcd_selectprofile(struct interface *ifp, const char *profile)
 }
 
 static void
-configure_interface(struct interface *ifp, int argc, char **argv)
+configure_interface(struct interface *ifp, int argc, char **argv,
+    unsigned long long options)
 {
 
 	dhcpcd_selectprofile(ifp, NULL);
 	add_options(ifp->ctx, ifp->name, ifp->options, argc, argv);
+	ifp->options->options |= options;
 	configure_interface1(ifp);
 }
 
@@ -548,10 +564,10 @@ dhcpcd_pollup(void *arg)
 
 	carrier = if_carrier(ifp); /* will set ifp->flags */
 	if (carrier == LINK_UP && !(ifp->flags & IFF_UP)) {
-		struct timeval tv;
+		struct timespec tv;
 
 		tv.tv_sec = 0;
-		tv.tv_usec = IF_POLL_UP * 1000;
+		tv.tv_nsec = IF_POLL_UP * MSEC_PER_NSEC;
 		eloop_timeout_add_tv(ifp->ctx->eloop, &tv, dhcpcd_pollup, ifp);
 		return;
 	}
@@ -594,11 +610,11 @@ dhcpcd_handlecarrier(struct dhcpcd_ctx *ctx, int carrier, unsigned int flags,
 
 	if (carrier == LINK_UNKNOWN) {
 		if (errno != ENOTTY) /* For example a PPP link on BSD */
-			syslog(LOG_ERR, "%s: carrier_status: %m", ifname);
+			logger(ctx, LOG_ERR, "%s: carrier_status: %m", ifname);
 	} else if (carrier == LINK_DOWN || (ifp->flags & IFF_UP) == 0) {
 		if (ifp->carrier != LINK_DOWN) {
 			if (ifp->carrier == LINK_UP)
-				syslog(LOG_INFO, "%s: carrier lost", ifp->name);
+				logger(ctx, LOG_INFO, "%s: carrier lost", ifp->name);
 			ifp->carrier = LINK_DOWN;
 			script_runreason(ifp, "NOCARRIER");
 			dhcp6_drop(ifp, "EXPIRE6");
@@ -609,7 +625,7 @@ dhcpcd_handlecarrier(struct dhcpcd_ctx *ctx, int carrier, unsigned int flags,
 		}
 	} else if (carrier == LINK_UP && ifp->flags & IFF_UP) {
 		if (ifp->carrier != LINK_UP) {
-			syslog(LOG_INFO, "%s: carrier acquired", ifp->name);
+			logger(ctx, LOG_INFO, "%s: carrier acquired", ifp->name);
 			ifp->carrier = LINK_UP;
 #if !defined(__linux__) && !defined(__NetBSD__)
 			/* BSD does not emit RTM_NEWADDR or RTM_CHGADDR when the
@@ -619,7 +635,7 @@ dhcpcd_handlecarrier(struct dhcpcd_ctx *ctx, int carrier, unsigned int flags,
 #endif
 			if (ifp->wireless)
 				if_getssid(ifp);
-			dhcpcd_initstate(ifp);
+			dhcpcd_initstate(ifp, 0);
 			script_runreason(ifp, "CARRIER");
 			/* RFC4941 Section 3.5 */
 			if (ifp->options->options & DHCPCD_IPV6RA_OWN)
@@ -652,7 +668,7 @@ warn_iaid_conflict(struct interface *ifp, uint8_t *iaid)
 
 	/* This is only a problem if the interfaces are on the same network. */
 	if (ifn && strcmp(ifp->name, ifn->name))
-		syslog(LOG_ERR,
+		logger(ifp->ctx, LOG_ERR,
 		    "%s: IAID conflicts with one assigned to %s",
 		    ifp->name, ifn->name);
 }
@@ -666,7 +682,7 @@ pre_start(struct interface *ifp)
 	 * This is also a safety check incase it was ripped out
 	 * from under us. */
 	if (ifp->options->options & DHCPCD_IPV6 && ipv6_start(ifp) == -1) {
-		syslog(LOG_ERR, "%s: ipv6_start: %m", ifp->name);
+		logger(ifp->ctx, LOG_ERR, "%s: ipv6_start: %m", ifp->name);
 		ifp->options->options &= ~DHCPCD_IPV6;
 	}
 }
@@ -679,21 +695,22 @@ dhcpcd_startinterface(void *arg)
 	size_t i;
 	char buf[DUID_LEN * 3];
 	int carrier;
-	struct timeval tv;
+	struct timespec tv;
 
 	if (ifo->options & DHCPCD_LINK) {
 		switch (ifp->carrier) {
 		case LINK_UP:
 			break;
 		case LINK_DOWN:
-			syslog(LOG_INFO, "%s: waiting for carrier", ifp->name);
+			logger(ifp->ctx, LOG_INFO, "%s: waiting for carrier",
+			    ifp->name);
 			return;
 		case LINK_UNKNOWN:
 			/* No media state available.
 			 * Loop until both IFF_UP and IFF_RUNNING are set */
 			if ((carrier = if_carrier(ifp)) == LINK_UNKNOWN) {
 				tv.tv_sec = 0;
-				tv.tv_usec = IF_POLL_UP * 1000;
+				tv.tv_nsec = IF_POLL_UP * MSEC_PER_NSEC;
 				eloop_timeout_add_tv(ifp->ctx->eloop,
 				    &tv, dhcpcd_startinterface, ifp);
 			} else
@@ -709,7 +726,7 @@ dhcpcd_startinterface(void *arg)
 			if (duid_init(ifp) == 0)
 				return;
 			if (!(ifo->options & DHCPCD_PFXDLGONLY))
-				syslog(LOG_INFO, "DUID %s",
+				logger(ifp->ctx, LOG_INFO, "DUID %s",
 				    hwaddr_ntoa(ifp->ctx->duid,
 				    ifp->ctx->duid_len,
 				    buf, sizeof(buf)));
@@ -720,7 +737,7 @@ dhcpcd_startinterface(void *arg)
 	    !(ifo->options & DHCPCD_PFXDLGONLY))
 	{
 		/* Report IAIDs */
-		syslog(LOG_INFO, "%s: IAID %s", ifp->name,
+		logger(ifp->ctx, LOG_INFO, "%s: IAID %s", ifp->name,
 		    hwaddr_ntoa(ifo->iaid, sizeof(ifo->iaid),
 		    buf, sizeof(buf)));
 		warn_iaid_conflict(ifp, ifo->iaid);
@@ -728,8 +745,8 @@ dhcpcd_startinterface(void *arg)
 			if (memcmp(ifo->iaid, ifo->ia[i].iaid,
 			    sizeof(ifo->iaid)))
 			{
-				syslog(LOG_INFO, "%s: IAID %s", ifp->name,
-				    hwaddr_ntoa(ifo->ia[i].iaid,
+				logger(ifp->ctx, LOG_INFO, "%s: IAID %s",
+				    ifp->name, hwaddr_ntoa(ifo->ia[i].iaid,
 				    sizeof(ifo->ia[i].iaid),
 				    buf, sizeof(buf)));
 				warn_iaid_conflict(ifp, ifo->ia[i].iaid);
@@ -739,7 +756,7 @@ dhcpcd_startinterface(void *arg)
 
 	if (ifo->options & DHCPCD_IPV6) {
 		if (ifo->options & DHCPCD_IPV6RS &&
-		    !(ifo->options & (DHCPCD_INFORM | DHCPCD_PFXDLGONLY)))
+		    !(ifo->options & DHCPCD_INFORM))
 			ipv6nd_startrs(ifp);
 
 		if (ifo->options & DHCPCD_DHCP6)
@@ -769,12 +786,10 @@ dhcpcd_startinterface(void *arg)
 #endif
 			}
 			if (nolease == -1)
-			        syslog(LOG_ERR,
+			        logger(ifp->ctx, LOG_ERR,
 				    "%s: dhcp6_start: %m", ifp->name);
 		}
 	}
-	if (ifo->options & DHCPCD_PFXDLGONLY)
-		return;
 
 	if (ifo->options & DHCPCD_IPV4)
 		dhcp_start(ifp);
@@ -787,7 +802,7 @@ dhcpcd_prestartinterface(void *arg)
 
 	pre_start(ifp);
 	if (if_up(ifp) == -1)
-		syslog(LOG_ERR, "%s: if_up: %m", ifp->name);
+		logger(ifp->ctx, LOG_ERR, "%s: if_up: %m", ifp->name);
 
 	if (ifp->options->options & DHCPCD_LINK &&
 	    ifp->carrier == LINK_UNKNOWN)
@@ -799,7 +814,7 @@ dhcpcd_prestartinterface(void *arg)
 			    ifp->flags, ifp->name);
 			return;
 		}
-		syslog(LOG_INFO,
+		logger(ifp->ctx, LOG_INFO,
 		    "%s: unknown carrier, waiting for interface flags",
 		    ifp->name);
 	}
@@ -814,7 +829,7 @@ handle_link(void *arg)
 
 	ctx = arg;
 	if (if_managelink(ctx) == -1) {
-		syslog(LOG_ERR, "if_managelink: %m");
+		logger(ctx, LOG_ERR, "if_managelink: %m");
 		eloop_event_delete(ctx->eloop, ctx->link_fd, 0);
 		close(ctx->link_fd);
 		ctx->link_fd = -1;
@@ -822,19 +837,20 @@ handle_link(void *arg)
 }
 
 static void
-dhcpcd_initstate1(struct interface *ifp, int argc, char **argv)
+dhcpcd_initstate1(struct interface *ifp, int argc, char **argv,
+    unsigned long long options)
 {
 	struct if_options *ifo;
 
-	configure_interface(ifp, argc, argv);
+	configure_interface(ifp, argc, argv, options);
 	ifo = ifp->options;
 
 	if (ifo->options & DHCPCD_IPV4 && ipv4_init(ifp->ctx) == -1) {
-		syslog(LOG_ERR, "ipv4_init: %m");
+		logger(ifp->ctx, LOG_ERR, "ipv4_init: %m");
 		ifo->options &= ~DHCPCD_IPV4;
 	}
 	if (ifo->options & DHCPCD_IPV6 && ipv6_init(ifp->ctx) == NULL) {
-		syslog(LOG_ERR, "ipv6_init: %m");
+		logger(ifp->ctx, LOG_ERR, "ipv6_init: %m");
 		ifo->options &= ~DHCPCD_IPV6RS;
 	}
 
@@ -843,16 +859,16 @@ dhcpcd_initstate1(struct interface *ifp, int argc, char **argv)
 	 * This needs to happen before PREINIT incase a hook script
 	 * inadvertently ups the interface. */
 	if (ifo->options & DHCPCD_IPV6 && ipv6_start(ifp) == -1) {
-		syslog(LOG_ERR, "%s: ipv6_start: %m", ifp->name);
+		logger(ifp->ctx, LOG_ERR, "%s: ipv6_start: %m", ifp->name);
 		ifo->options &= ~DHCPCD_IPV6;
 	}
 }
 
 void
-dhcpcd_initstate(struct interface *ifp)
+dhcpcd_initstate(struct interface *ifp, unsigned long long options)
 {
 
-	dhcpcd_initstate1(ifp, ifp->ctx->argc, ifp->ctx->argv);
+	dhcpcd_initstate1(ifp, ifp->ctx->argc, ifp->ctx->argv, options);
 }
 
 static void
@@ -886,7 +902,7 @@ dhcpcd_handleinterface(void *arg, int action, const char *ifname)
 			errno = ESRCH;
 			return -1;
 		}
-		syslog(LOG_DEBUG, "%s: interface departed", ifp->name);
+		logger(ctx, LOG_DEBUG, "%s: interface departed", ifp->name);
 		ifp->options->options |= DHCPCD_DEPARTED;
 		stop_interface(ifp);
 		return 0;
@@ -904,7 +920,7 @@ dhcpcd_handleinterface(void *arg, int action, const char *ifname)
 	i = -1;
 	ifs = if_discover(ctx, -1, UNCONST(argv));
 	if (ifs == NULL) {
-		syslog(LOG_ERR, "%s: if_discover: %m", __func__);
+		logger(ctx, LOG_ERR, "%s: if_discover: %m", __func__);
 		return -1;
 	}
 	TAILQ_FOREACH_SAFE(ifp, ifs, next, ifn) {
@@ -914,17 +930,17 @@ dhcpcd_handleinterface(void *arg, int action, const char *ifname)
 		/* Check if we already have the interface */
 		iff = if_find(ctx, ifp->name);
 		if (iff) {
-			syslog(LOG_DEBUG, "%s: interface updated", iff->name);
+			logger(ctx, LOG_DEBUG, "%s: interface updated", iff->name);
 			/* The flags and hwaddr could have changed */
 			iff->flags = ifp->flags;
 			iff->hwlen = ifp->hwlen;
 			if (ifp->hwlen != 0)
 				memcpy(iff->hwaddr, ifp->hwaddr, iff->hwlen);
 		} else {
-			syslog(LOG_DEBUG, "%s: interface added", ifp->name);
+			logger(ctx, LOG_DEBUG, "%s: interface added", ifp->name);
 			TAILQ_REMOVE(ifs, ifp, next);
 			TAILQ_INSERT_TAIL(ctx->ifaces, ifp, next);
-			dhcpcd_initstate(ifp);
+			dhcpcd_initstate(ifp, 0);
 			run_preinit(ifp);
 			iff = ifp;
 		}
@@ -957,14 +973,14 @@ dhcpcd_handlehwaddr(struct dhcpcd_ctx *ctx, const char *ifname,
 
 	if (hwlen > sizeof(ifp->hwaddr)) {
 		errno = ENOBUFS;
-		syslog(LOG_ERR, "%s: %s: %m", ifp->name, __func__);
+		logger(ctx, LOG_ERR, "%s: %s: %m", ifp->name, __func__);
 		return;
 	}
 
 	if (ifp->hwlen == hwlen && memcmp(ifp->hwaddr, hwaddr, hwlen) == 0)
 		return;
 
-	syslog(LOG_INFO, "%s: new hardware address: %s", ifp->name,
+	logger(ctx, LOG_INFO, "%s: new hardware address: %s", ifp->name,
 	    hwaddr_ntoa(hwaddr, hwlen, buf, sizeof(buf)));
 	ifp->hwlen = hwlen;
 	memcpy(ifp->hwaddr, hwaddr, hwlen);
@@ -977,7 +993,7 @@ if_reboot(struct interface *ifp, int argc, char **argv)
 
 	oldopts = ifp->options->options;
 	script_runreason(ifp, "RECONFIGURE");
-	configure_interface(ifp, argc, argv);
+	dhcpcd_initstate1(ifp, argc, argv, 0);
 	dhcp_reboot_newopts(ifp, oldopts);
 	dhcp6_reboot(ifp);
 	dhcpcd_prestartinterface(ifp);
@@ -1008,7 +1024,7 @@ reconf_reboot(struct dhcpcd_ctx *ctx, int action, int argc, char **argv, int oi)
 
 	ifs = if_discover(ctx, argc - oi, argv + oi);
 	if (ifs == NULL) {
-		syslog(LOG_ERR, "%s: if_discover: %m", __func__);
+		logger(ctx, LOG_ERR, "%s: if_discover: %m", __func__);
 		return;
 	}
 
@@ -1023,14 +1039,12 @@ reconf_reboot(struct dhcpcd_ctx *ctx, int action, int argc, char **argv, int oi)
 			if_free(ifp);
 		} else {
 			TAILQ_INSERT_TAIL(ctx->ifaces, ifp, next);
-			dhcpcd_initstate1(ifp, argc, argv);
+			dhcpcd_initstate1(ifp, argc, argv, 0);
 			run_preinit(ifp);
 			dhcpcd_prestartinterface(ifp);
 		}
 	}
 	free(ifs);
-
-	ipv4_sortinterfaces(ctx);
 }
 
 static void
@@ -1058,36 +1072,35 @@ stop_all_interfaces(struct dhcpcd_ctx *ctx, int do_release)
 }
 
 #ifdef USE_SIGNALS
-struct dhcpcd_siginfo {
-	int signo;
-	pid_t pid;
-} dhcpcd_siginfo;
-
-#define sigmsg "received signal %s from PID %d, %s"
-static void
-handle_signal1(void *arg)
+struct dhcpcd_siginfo dhcpcd_siginfo;
+#define sigmsg "received %s, %s"
+void
+dhcpcd_handle_signal(void *arg)
 {
 	struct dhcpcd_ctx *ctx;
 	struct dhcpcd_siginfo *si;
 	struct interface *ifp;
-	int do_release;
+	int do_release, exit_code;;
 
 	ctx = dhcpcd_ctx;
 	si = arg;
 	do_release = 0;
+	exit_code = EXIT_FAILURE;
 	switch (si->signo) {
 	case SIGINT:
-		syslog(LOG_INFO, sigmsg, "INT", (int)si->pid, "stopping");
+		logger(ctx, LOG_INFO, sigmsg, "SIGINT", "stopping");
 		break;
 	case SIGTERM:
-		syslog(LOG_INFO, sigmsg, "TERM", (int)si->pid, "stopping");
+		logger(ctx, LOG_INFO, sigmsg, "SIGTERM", "stopping");
+		exit_code = EXIT_SUCCESS;
 		break;
 	case SIGALRM:
-		syslog(LOG_INFO, sigmsg, "ALRM", (int)si->pid, "releasing");
+		logger(ctx, LOG_INFO, sigmsg, "SIGALRM", "releasing");
 		do_release = 1;
+		exit_code = EXIT_SUCCESS;
 		break;
 	case SIGHUP:
-		syslog(LOG_INFO, sigmsg, "HUP", (int)si->pid, "rebinding");
+		logger(ctx, LOG_INFO, sigmsg, "SIGHUP", "rebinding");
 		reload_config(ctx);
 		/* Preserve any options passed on the commandline
 		 * when we were started. */
@@ -1095,51 +1108,60 @@ handle_signal1(void *arg)
 		    ctx->argc - ctx->ifc);
 		return;
 	case SIGUSR1:
-		syslog(LOG_INFO, sigmsg, "USR1", (int)si->pid, "reconfiguring");
+		logger(ctx, LOG_INFO, sigmsg, "SIGUSR1", "reconfiguring");
 		TAILQ_FOREACH(ifp, ctx->ifaces, next) {
 			ipv4_applyaddr(ifp);
 		}
 		return;
+	case SIGUSR2:
+		logger_close(ctx);
+		logger_open(ctx);
+		logger(ctx, LOG_INFO, sigmsg, "SIGUSR2", "reopened logfile");
+		return;
 	case SIGPIPE:
-		syslog(LOG_WARNING, "received signal PIPE");
+		logger(ctx, LOG_WARNING, "received SIGPIPE");
 		return;
 	default:
-		syslog(LOG_ERR,
-		    "received signal %d from PID %d, "
+		logger(ctx, LOG_ERR,
+		    "received signal %d, "
 		    "but don't know what to do with it",
-		    si->signo, (int)si->pid);
+		    si->signo);
 		return;
 	}
 
 	if (!(ctx->options & DHCPCD_TEST))
 		stop_all_interfaces(ctx, do_release);
-	eloop_exit(ctx->eloop, EXIT_FAILURE);
+	eloop_exit(ctx->eloop, exit_code);
 }
 
+#ifndef HAVE_KQUEUE
 static void
-handle_signal(int sig, siginfo_t *siginfo, __unused void *context)
+handle_signal(int sig, __unused siginfo_t *siginfo, __unused void *context)
 {
 
 	/* So that we can operate safely under a signal we instruct
 	 * eloop to pass a copy of the siginfo structure to handle_signal1
 	 * as the very first thing to do. */
 	dhcpcd_siginfo.signo = sig;
-	dhcpcd_siginfo.pid = siginfo ? siginfo->si_pid : 0;
 	eloop_timeout_add_now(dhcpcd_ctx->eloop,
-	    handle_signal1, &dhcpcd_siginfo);
+	    dhcpcd_handle_signal, &dhcpcd_siginfo);
 }
+#endif
 
 static int
 signal_init(sigset_t *oldset)
 {
-	unsigned int i;
-	struct sigaction sa;
 	sigset_t newset;
+#ifndef HAVE_KQUEUE
+	int i;
+	struct sigaction sa;
+#endif
 
 	sigfillset(&newset);
 	if (sigprocmask(SIG_SETMASK, &newset, oldset) == -1)
 		return -1;
 
+#ifndef HAVE_KQUEUE
 	memset(&sa, 0, sizeof(sa));
 	sa.sa_sigaction = handle_signal;
 	sa.sa_flags = SA_SIGINFO;
@@ -1149,6 +1171,7 @@ signal_init(sigset_t *oldset)
 		if (sigaction(dhcpcd_handlesigs[i], &sa, NULL) == -1)
 			return -1;
 	}
+#endif
 	return 0;
 }
 #endif
@@ -1175,7 +1198,8 @@ dhcpcd_getinterfaces(void *arg)
 	eloop_event_delete(fd->ctx->eloop, fd->fd, 1);
 	TAILQ_FOREACH(ifp, fd->ctx->ifaces, next) {
 		if (send_interface(fd, ifp) == -1)
-			syslog(LOG_ERR, "send_interface %d: %m", fd->fd);
+			logger(ifp->ctx, LOG_ERR,
+			    "send_interface %d: %m", fd->fd);
 	}
 }
 
@@ -1230,7 +1254,7 @@ dhcpcd_handleargs(struct dhcpcd_ctx *ctx, struct fd_list *fd,
 		*p++ = ' ';
 	}
 	*--p = '\0';
-	syslog(LOG_INFO, "control command: %s", tmp);
+	logger(ctx, LOG_INFO, "control command: %s", tmp);
 	free(tmp);
 
 	optind = 0;
@@ -1281,7 +1305,6 @@ int
 main(int argc, char **argv)
 {
 	struct dhcpcd_ctx ctx;
-	char pidfile[sizeof(PIDFILE) + IF_NAMESIZE + 1];
 	struct if_options *ifo;
 	struct interface *ifp;
 	uint16_t family = 0;
@@ -1295,24 +1318,6 @@ main(int argc, char **argv)
 	int sig;
 	const char *siga;
 #endif
-	struct timespec ts;
-
-	memset(&ctx, 0, sizeof(ctx));
-#ifdef USE_SIGNALS
-	dhcpcd_ctx = &ctx;
-	sig = 0;
-	siga = NULL;
-#endif
-	closefrom(3);
-#ifdef LOG_PERROR
-	openlog(PACKAGE, LOG_PERROR | LOG_PID, LOG_DAEMON);
-#else
-	openlog(PACKAGE, LOG_PID, LOG_DAEMON);
-#endif
-	setlogmask(LOG_UPTO(LOG_INFO));
-#ifndef LOG_PERROR
-	psyslog_prio = LOG_UPTO(LOG_INFO);
-#endif
 
 	/* Test for --help and --version */
 	if (argc > 1) {
@@ -1325,6 +1330,18 @@ main(int argc, char **argv)
 		}
 	}
 
+	memset(&ctx, 0, sizeof(ctx));
+#ifdef USE_SIGNALS
+	dhcpcd_ctx = &ctx;
+	sig = 0;
+	siga = NULL;
+#endif
+	closefrom(3);
+
+	ctx.log_fd = -1;
+	logger_open(&ctx);
+	logger_mask(&ctx, LOG_UPTO(LOG_INFO));
+
 	ifo = NULL;
 	ctx.cffile = CONFIG;
 	ctx.pid_fd = ctx.control_fd = ctx.control_unpriv_fd = ctx.link_fd = -1;
@@ -1340,19 +1357,9 @@ main(int argc, char **argv)
 	{
 		switch (opt) {
 		case '4':
-			if (family) {
-				syslog(LOG_ERR, "cannot specify more than one"
-				    " address family");
-				goto exit_failure;
-			}
 			family = AF_INET;
 			break;
 		case '6':
-			if (family) {
-				syslog(LOG_ERR, "cannot specify more than one"
-				    " address family");
-				goto exit_failure;
-			}
 			family = AF_INET6;
 			break;
 		case 'f':
@@ -1363,6 +1370,11 @@ main(int argc, char **argv)
 			sig = SIGUSR1;
 			siga = "USR1";
 			break;
+		case 'j':
+			ctx.logfile = strdup(optarg);
+			logger_close(&ctx);
+			logger_open(&ctx);
+			break;
 		case 'k':
 			sig = SIGALRM;
 			siga = "ARLM";
@@ -1450,16 +1462,12 @@ main(int argc, char **argv)
 	ctx.options &= ~DHCPCD_DAEMONISE;
 #endif
 
-	if (ctx.options & DHCPCD_DEBUG) {
-		setlogmask(LOG_UPTO(LOG_DEBUG));
-#ifndef LOG_PERROR
-		psyslog_prio = LOG_UPTO(LOG_DEBUG);
-#endif
-	}
+	if (ctx.options & DHCPCD_DEBUG)
+		logger_mask(&ctx, LOG_UPTO(LOG_DEBUG));
 	if (ctx.options & DHCPCD_QUIET) {
 		i = open(_PATH_DEVNULL, O_RDWR);
 		if (i == -1)
-			syslog(LOG_ERR, "%s: open: %m", __func__);
+			logger(&ctx, LOG_ERR, "%s: open: %m", __func__);
 		else {
 			dup2(i, STDERR_FILENO);
 			close(i);
@@ -1473,7 +1481,8 @@ main(int argc, char **argv)
 			const char *per;
 
 			if (strlen(argv[optind]) > IF_NAMESIZE) {
-				syslog(LOG_ERR, "%s: interface name too long",
+				logger(&ctx, LOG_ERR,
+				    "%s: interface name too long",
 				    argv[optind]);
 				goto exit_failure;
 			}
@@ -1488,28 +1497,30 @@ main(int argc, char **argv)
 			default:
 				per = "";
 			}
-			snprintf(pidfile, sizeof(pidfile),
+			snprintf(ctx.pidfile, sizeof(ctx.pidfile),
 			    PIDFILE, "-", argv[optind], per);
 		} else {
-			snprintf(pidfile, sizeof(pidfile), PIDFILE, "", "", "");
+			snprintf(ctx.pidfile, sizeof(ctx.pidfile),
+			    PIDFILE, "", "", "");
 			ctx.options |= DHCPCD_MASTER;
 		}
 	}
 
 	if (chdir("/") == -1)
-		syslog(LOG_ERR, "chdir `/': %m");
+		logger(&ctx, LOG_ERR, "chdir `/': %m");
 
 	/* Freeing allocated addresses from dumping leases can trigger
 	 * eloop removals as well, so init here. */
-	ctx.eloop = eloop_init();
+	ctx.eloop = eloop_init(&ctx);
 	if (ctx.eloop == NULL) {
-		syslog(LOG_ERR, "%s: %m", __func__);
+		logger(&ctx, LOG_ERR, "%s: eloop_init: %m", __func__);
 		goto exit_failure;
 	}
 
 	if (ctx.options & DHCPCD_DUMPLEASE) {
 		if (optind != argc - 1) {
-			syslog(LOG_ERR, "dumplease requires an interface");
+			logger(&ctx, LOG_ERR,
+			    "dumplease requires an interface");
 			goto exit_failure;
 		}
 		i = 0;
@@ -1517,21 +1528,27 @@ main(int argc, char **argv)
 		 * load the hardware address to compare automated IAID */
 		ctx.ifaces = if_discover(&ctx, 1, argv + optind);
 		if (ctx.ifaces == NULL) {
-			syslog(LOG_ERR, "if_discover: %m");
+			logger(&ctx, LOG_ERR, "if_discover: %m");
 			goto exit_failure;
 		}
 		ifp = TAILQ_FIRST(ctx.ifaces);
 		if (ifp == NULL) {
 			ifp = calloc(1, sizeof(*ifp));
 			if (ifp == NULL) {
-				syslog(LOG_ERR, "%s: %m", __func__);
+				logger(&ctx, LOG_ERR, "%s: %m", __func__);
 				goto exit_failure;
 			}
-			strlcpy(ifp->name, argv[optind], sizeof(ifp->name));
+			strlcpy(ctx.pidfile, argv[optind], sizeof(ctx.pidfile));
 			ifp->ctx = &ctx;
 			TAILQ_INSERT_HEAD(ctx.ifaces, ifp, next);
+			if (family == 0) {
+				if (ctx.pidfile[strlen(ctx.pidfile) - 1] == '6')
+					family = AF_INET6;
+				else
+					family = AF_INET;
+			}
 		}
-		configure_interface(ifp, ctx.argc, ctx.argv);
+		configure_interface(ifp, ctx.argc, ctx.argv, 0);
 		if (ctx.options & DHCPCD_PFXDLGONLY)
 			ifp->options->options |= DHCPCD_PFXDLGONLY;
 		if (family == 0 || family == AF_INET) {
@@ -1559,91 +1576,95 @@ main(int argc, char **argv)
 		if (i == -1)
 			i = control_open(&ctx, NULL);
 		if (i != -1) {
-			syslog(LOG_INFO,
+			logger(&ctx, LOG_INFO,
 			    "sending commands to master dhcpcd process");
 			len = control_send(&ctx, argc, argv);
 			control_close(&ctx);
 			if (len > 0) {
-				syslog(LOG_DEBUG, "send OK");
+				logger(&ctx, LOG_DEBUG, "send OK");
 				goto exit_success;
 			} else {
-				syslog(LOG_ERR, "failed to send commands");
+				logger(&ctx, LOG_ERR,
+				    "failed to send commands");
 				goto exit_failure;
 			}
 		} else {
 			if (errno != ENOENT)
-				syslog(LOG_ERR, "control_open: %m");
+				logger(&ctx, LOG_ERR, "control_open: %m");
 		}
 #ifdef USE_SIGNALS
 	}
 #endif
 
 	if (geteuid())
-		syslog(LOG_WARNING,
+		logger(&ctx, LOG_WARNING,
 		    PACKAGE " will not work correctly unless run as root");
 
 #ifdef USE_SIGNALS
 	if (sig != 0) {
-		pid = read_pid(pidfile);
+		pid = read_pid(ctx.pidfile);
 		if (pid != 0)
-			syslog(LOG_INFO, "sending signal %s to pid %d",
+			logger(&ctx, LOG_INFO, "sending signal %s to pid %d",
 			    siga, pid);
 		if (pid == 0 || kill(pid, sig) != 0) {
 			if (sig != SIGHUP && errno != EPERM)
-				syslog(LOG_ERR, ""PACKAGE" not running");
+				logger(&ctx, LOG_ERR, ""PACKAGE" not running");
 			if (pid != 0 && errno != ESRCH) {
-				syslog(LOG_ERR, "kill: %m");
+				logger(&ctx, LOG_ERR, "kill: %m");
 				goto exit_failure;
 			}
-			unlink(pidfile);
+			unlink(ctx.pidfile);
 			if (sig != SIGHUP)
 				goto exit_failure;
 		} else {
+			struct timespec ts;
+
 			if (sig == SIGHUP || sig == SIGUSR1)
 				goto exit_success;
 			/* Spin until it exits */
-			syslog(LOG_INFO, "waiting for pid %d to exit", pid);
+			logger(&ctx, LOG_INFO,
+			    "waiting for pid %d to exit", pid);
 			ts.tv_sec = 0;
 			ts.tv_nsec = 100000000; /* 10th of a second */
 			for(i = 0; i < 100; i++) {
 				nanosleep(&ts, NULL);
-				if (read_pid(pidfile) == 0)
+				if (read_pid(ctx.pidfile) == 0)
 					goto exit_success;
 			}
-			syslog(LOG_ERR, "pid %d failed to exit", pid);
+			logger(&ctx, LOG_ERR, "pid %d failed to exit", pid);
 			goto exit_failure;
 		}
 	}
 
 	if (!(ctx.options & DHCPCD_TEST)) {
-		if ((pid = read_pid(pidfile)) > 0 &&
+		if ((pid = read_pid(ctx.pidfile)) > 0 &&
 		    kill(pid, 0) == 0)
 		{
-			syslog(LOG_ERR, ""PACKAGE
+			logger(&ctx, LOG_ERR, ""PACKAGE
 			    " already running on pid %d (%s)",
-			    pid, pidfile);
+			    pid, ctx.pidfile);
 			goto exit_failure;
 		}
 
 		/* Ensure we have the needed directories */
 		if (mkdir(RUNDIR, 0755) == -1 && errno != EEXIST)
-			syslog(LOG_ERR, "mkdir `%s': %m", RUNDIR);
+			logger(&ctx, LOG_ERR, "mkdir `%s': %m", RUNDIR);
 		if (mkdir(DBDIR, 0755) == -1 && errno != EEXIST)
-			syslog(LOG_ERR, "mkdir `%s': %m", DBDIR);
+			logger(&ctx, LOG_ERR, "mkdir `%s': %m", DBDIR);
 
 		opt = O_WRONLY | O_CREAT | O_NONBLOCK;
 #ifdef O_CLOEXEC
 		opt |= O_CLOEXEC;
 #endif
-		ctx.pid_fd = open(pidfile, opt, 0664);
+		ctx.pid_fd = open(ctx.pidfile, opt, 0664);
 		if (ctx.pid_fd == -1)
-			syslog(LOG_ERR, "open `%s': %m", pidfile);
+			logger(&ctx, LOG_ERR, "open `%s': %m", ctx.pidfile);
 		else {
 #ifdef LOCK_EX
 			/* Lock the file so that only one instance of dhcpcd
 			 * runs on an interface */
 			if (flock(ctx.pid_fd, LOCK_EX | LOCK_NB) == -1) {
-				syslog(LOG_ERR, "flock `%s': %m", pidfile);
+				logger(&ctx, LOG_ERR, "flock `%s': %m", ctx.pidfile);
 				close(ctx.pid_fd);
 				ctx.pid_fd = -1;
 				goto exit_failure;
@@ -1653,7 +1674,7 @@ main(int argc, char **argv)
 			if (fcntl(ctx.pid_fd, F_GETFD, &opt) == -1 ||
 			    fcntl(ctx.pid_fd, F_SETFD, opt | FD_CLOEXEC) == -1)
 			{
-				syslog(LOG_ERR, "fcntl: %m");
+				logger(&ctx, LOG_ERR, "fcntl: %m");
 				close(ctx.pid_fd);
 				ctx.pid_fd = -1;
 				goto exit_failure;
@@ -1666,23 +1687,23 @@ main(int argc, char **argv)
 
 	if (ctx.options & DHCPCD_MASTER) {
 		if (control_start(&ctx, NULL) == -1)
-			syslog(LOG_ERR, "control_start: %m");
+			logger(&ctx, LOG_ERR, "control_start: %m");
 	}
 #else
 	if (control_start(&ctx,
 	    ctx.options & DHCPCD_MASTER ? NULL : argv[optind]) == -1)
 	{
-		syslog(LOG_ERR, "control_start: %m");
+		logger(&ctx, LOG_ERR, "control_start: %m");
 		goto exit_failure;
 	}
 #endif
 
-	syslog(LOG_INFO, "version " VERSION " starting");
+	logger(&ctx, LOG_DEBUG, PACKAGE "-" VERSION " starting");
 	ctx.options |= DHCPCD_STARTED;
 #ifdef USE_SIGNALS
 	/* Save signal mask, block and redirect signals to our handler */
 	if (signal_init(&ctx.sigset) == -1) {
-		syslog(LOG_ERR, "signal_setup: %m");
+		logger(&ctx, LOG_ERR, "signal_setup: %m");
 		goto exit_failure;
 	}
 #endif
@@ -1699,7 +1720,7 @@ main(int argc, char **argv)
 	 * so that we pickup any new addresses during the discover phase. */
 	ctx.link_fd = if_openlinksocket();
 	if (ctx.link_fd == -1)
-		syslog(LOG_ERR, "open_link_socket: %m");
+		logger(&ctx, LOG_ERR, "open_link_socket: %m");
 	else
 		eloop_event_add(ctx.eloop, ctx.link_fd,
 		    handle_link, &ctx, NULL, NULL);
@@ -1712,28 +1733,29 @@ main(int argc, char **argv)
 
 	ctx.ifaces = if_discover(&ctx, ctx.ifc, ctx.ifv);
 	if (ctx.ifaces == NULL) {
-		syslog(LOG_ERR, "if_discover: %m");
+		logger(&ctx, LOG_ERR, "if_discover: %m");
 		goto exit_failure;
 	}
 	for (i = 0; i < ctx.ifc; i++) {
 		if (if_find(&ctx, ctx.ifv[i]) == NULL)
-			syslog(LOG_ERR, "%s: interface not found or invalid",
+			logger(&ctx, LOG_ERR,
+			    "%s: interface not found or invalid",
 			    ctx.ifv[i]);
 	}
 	if (TAILQ_FIRST(ctx.ifaces) == NULL) {
 		if (ctx.ifc == 0)
-			syslog(LOG_ERR, "no valid interfaces found");
+			logger(&ctx, LOG_ERR, "no valid interfaces found");
 		else
 			goto exit_failure;
 		if (!(ctx.options & DHCPCD_LINK)) {
-			syslog(LOG_ERR,
+			logger(&ctx, LOG_ERR,
 			    "aborting as link detection is disabled");
 			goto exit_failure;
 		}
 	}
 
 	TAILQ_FOREACH(ifp, ctx.ifaces, next) {
-		dhcpcd_initstate1(ifp, argc, argv);
+		dhcpcd_initstate1(ifp, argc, argv, 0);
 	}
 
 	if (ctx.options & DHCPCD_BACKGROUND && dhcpcd_daemonise(&ctx))
@@ -1757,10 +1779,14 @@ main(int argc, char **argv)
 		    ctx.options & DHCPCD_LINK &&
 		    !(ctx.options & DHCPCD_WAITIP))
 		{
-			syslog(LOG_WARNING, "no interfaces have a carrier");
+			logger(&ctx, LOG_WARNING,
+			    "no interfaces have a carrier");
 			if (dhcpcd_daemonise(&ctx))
 				goto exit_success;
-		} else if (t > 0 && ctx.options & DHCPCD_DAEMONISE) {
+		} else if (t > 0 &&
+		    /* Test mode removes the daemonise bit, so check for both */
+		    ctx.options & (DHCPCD_DAEMONISE | DHCPCD_TEST))
+		{
 			eloop_timeout_add_sec(ctx.eloop, t,
 			    handle_exit_timeout, &ctx);
 		}
@@ -1768,13 +1794,13 @@ main(int argc, char **argv)
 	free_options(ifo);
 	ifo = NULL;
 
-	ipv4_sortinterfaces(&ctx);
+	if_sortinterfaces(&ctx);
 	TAILQ_FOREACH(ifp, ctx.ifaces, next) {
 		eloop_timeout_add_sec(ctx.eloop, 0,
 		    dhcpcd_prestartinterface, ifp);
 	}
 
-	i = eloop_start(&ctx);
+	i = eloop_start(ctx.eloop);
 	goto exit1;
 
 exit_success:
@@ -1805,15 +1831,16 @@ exit1:
 	ipv6_ctxfree(&ctx);
 	dev_stop(&ctx);
 	if (control_stop(&ctx) == -1)
-		syslog(LOG_ERR, "control_stop: %m:");
+		logger(&ctx, LOG_ERR, "control_stop: %m:");
 	if (ctx.pid_fd != -1) {
 		close(ctx.pid_fd);
-		unlink(pidfile);
+		unlink(ctx.pidfile);
 	}
 	eloop_free(ctx.eloop);
 
 	if (ctx.options & DHCPCD_STARTED && !(ctx.options & DHCPCD_FORKED))
-		syslog(LOG_INFO, "exited");
-	closelog();
+		logger(&ctx, LOG_INFO, PACKAGE " exited");
+	logger_close(&ctx);
+	free(ctx.logfile);
 	return i;
 }
