@@ -71,7 +71,6 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <syslog.h>
 #include <time.h>
 #include <unistd.h>
 
@@ -320,7 +319,7 @@ err_netlink(struct nlmsghdr *nlm)
 
 static int
 get_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp, int fd, int flags,
-    int (*callback)(struct dhcpcd_ctx *, struct interface *,struct nlmsghdr *))
+    int (*callback)(struct dhcpcd_ctx *, struct interface *, struct nlmsghdr *))
 {
 	char *buf = NULL, *nbuf;
 	ssize_t bytes;
@@ -363,8 +362,10 @@ get_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp, int fd, int flags,
 			goto eexit;
 		}
 		/* Ignore message if it is not from kernel */
-		if (nladdr.nl_pid != 0)
+		if (nladdr.nl_pid != 0) {
+			r = 0;
 			continue;
+		}
 
 		for (nlm = (struct nlmsghdr *)(void *)buf;
 		     nlm && NLMSG_OK(nlm, (size_t)bytes);
@@ -388,6 +389,116 @@ eexit:
 	return r;
 }
 
+#ifdef INET
+static int
+if_copyrt(struct dhcpcd_ctx *ctx, struct rt *rt, struct nlmsghdr *nlm)
+{
+	size_t len;
+	struct rtmsg *rtm;
+	struct rtattr *rta;
+	struct in_addr prefsrc;
+
+	len = nlm->nlmsg_len - sizeof(*nlm);
+	if (len < sizeof(*rtm)) {
+		errno = EBADMSG;
+		return -1;
+	}
+	rtm = (struct rtmsg *)NLMSG_DATA(nlm);
+	if (rtm->rtm_table != RT_TABLE_MAIN || rtm->rtm_family != AF_INET)
+		return -1;
+
+	memset(rt, 0, sizeof(*rt));
+	prefsrc.s_addr = INADDR_ANY;
+	rta = (struct rtattr *)RTM_RTA(rtm);
+	len = RTM_PAYLOAD(nlm);
+	while (RTA_OK(rta, len)) {
+		switch (rta->rta_type) {
+		case RTA_DST:
+			memcpy(&rt->dest.s_addr, RTA_DATA(rta),
+			    sizeof(rt->dest.s_addr));
+			break;
+		case RTA_GATEWAY:
+			memcpy(&rt->gate.s_addr, RTA_DATA(rta),
+			    sizeof(rt->gate.s_addr));
+			break;
+		case RTA_PREFSRC:
+			memcpy(&prefsrc.s_addr, RTA_DATA(rta),
+			    sizeof(prefsrc.s_addr));
+			break;
+		case RTA_OIF:
+			rt->iface = if_findindex(ctx,
+			    *(unsigned int *)RTA_DATA(rta));
+			break;
+		case RTA_PRIORITY:
+			rt->metric = *(unsigned int *)RTA_DATA(rta);
+			break;
+		}
+		rta = RTA_NEXT(rta, len);
+	}
+
+	inet_cidrtoaddr(rtm->rtm_dst_len, &rt->net);
+	if (rt->iface == NULL && prefsrc.s_addr != INADDR_ANY) {
+		struct ipv4_addr *ap;
+
+		/* For some reason the default route comes back with the
+		 * loopback interface in RTA_OIF? Lets find it by
+		 * preferred source address */
+		if ((ap = ipv4_findaddr(ctx, &prefsrc)))
+			rt->iface = ap->iface;
+	}
+	return 0;
+}
+#endif
+
+#ifdef INET6
+static int
+if_copyrt6(struct dhcpcd_ctx *ctx, struct rt6 *rt, struct nlmsghdr *nlm)
+{
+	size_t len;
+	struct rtmsg *rtm;
+	struct rtattr *rta;
+
+	len = nlm->nlmsg_len - sizeof(*nlm);
+	if (len < sizeof(*rtm)) {
+		errno = EBADMSG;
+		return -1;
+	}
+	rtm = (struct rtmsg *)NLMSG_DATA(nlm);
+	if (rtm->rtm_table != RT_TABLE_MAIN || rtm->rtm_family != AF_INET6)
+		return -1;
+
+	memset(rt, 0, sizeof(*rt));
+	ipv6_mask(&rt->net, rtm->rtm_dst_len);
+	if (rtm->rtm_type == RTN_UNREACHABLE)
+		rt->flags = RTF_REJECT;
+
+	rta = (struct rtattr *)RTM_RTA(rtm);
+	len = RTM_PAYLOAD(nlm);
+	while (RTA_OK(rta, len)) {
+		switch (rta->rta_type) {
+		case RTA_DST:
+			memcpy(&rt->dest.s6_addr, RTA_DATA(rta),
+			    sizeof(rt->dest.s6_addr));
+			break;
+		case RTA_GATEWAY:
+			memcpy(&rt->gate.s6_addr, RTA_DATA(rta),
+			    sizeof(rt->gate.s6_addr));
+			break;
+		case RTA_OIF:
+			rt->iface = if_findindex(ctx,
+			    *(unsigned int *)RTA_DATA(rta));
+			break;
+		case RTA_PRIORITY:
+			rt->metric = *(unsigned int *)RTA_DATA(rta);
+			break;
+		}
+		rta = RTA_NEXT(rta, len);
+	}
+
+	return 0;
+}
+#endif
+
 /* Work out the maximum pid size */
 static inline long long
 get_max_pid_t()
@@ -405,29 +516,31 @@ link_route(struct dhcpcd_ctx *ctx, __unused struct interface *ifp,
     struct nlmsghdr *nlm)
 {
 	size_t len;
-	unsigned int metric;
-	struct rtattr *rta;
 	struct rtmsg *rtm;
+	int cmd;
 #ifdef INET
 	struct rt rt;
 #endif
 #ifdef INET6
 	struct rt6 rt6;
 #endif
-
-	if (nlm->nlmsg_type != RTM_DELROUTE)
+	switch (nlm->nlmsg_type) {
+	case RTM_NEWROUTE:
+		cmd = RTM_ADD;
+		break;
+	case RTM_DELROUTE:
+		cmd = RTM_DELETE;
+		break;
+	default:
 		return 0;
+	}
 
 	len = nlm->nlmsg_len - sizeof(*nlm);
 	if (len < sizeof(*rtm)) {
 		errno = EBADMSG;
 		return -1;
 	}
-	rtm = NLMSG_DATA(nlm);
-	if (rtm->rtm_type != RTN_UNICAST ||
-	    rtm->rtm_table != RT_TABLE_MAIN ||
-	    (rtm->rtm_family != AF_INET && rtm->rtm_family != AF_INET6))
-		return 1;
+
 	/* Ignore messages generated by us.
 	 * For some reason we get messages generated by us
 	 * with a very large value in nlmsg_pid that seems to be
@@ -435,83 +548,23 @@ link_route(struct dhcpcd_ctx *ctx, __unused struct interface *ifp,
 	if (nlm->nlmsg_pid > get_max_pid_t())
 		return 1;
 
-	rta = (struct rtattr *)(void *)((char *)rtm +NLMSG_ALIGN(sizeof(*rtm)));
-	len = NLMSG_PAYLOAD(nlm, sizeof(*rtm));
-#ifdef INET
-	if (rtm->rtm_family == AF_INET)
-		memset(&rt, 0, sizeof(rt));
-#endif
-#ifdef INET6
-	if (rtm->rtm_family == AF_INET6)
-		memset(&rt6, 0, sizeof(rt6));
-#endif
-	metric = 0;
-	while (RTA_OK(rta, len)) {
-		switch (rtm->rtm_family) {
-#ifdef INET
-		case AF_INET:
-			switch (rta->rta_type) {
-			case RTA_DST:
-				memcpy(&rt.dest.s_addr, RTA_DATA(rta),
-				    sizeof(rt.dest.s_addr));
-				break;
-			case RTA_GATEWAY:
-				memcpy(&rt.gate.s_addr, RTA_DATA(rta),
-				    sizeof(rt.gate.s_addr));
-				break;
-			case RTA_OIF:
-				rt.iface = if_findindex(ctx,
-				    *(unsigned int *)RTA_DATA(rta));
-				break;
-			}
-			break;
-#endif
-#ifdef INET6
-		case AF_INET6:
-			switch (rta->rta_type) {
-			case RTA_DST:
-				memcpy(&rt6.dest.s6_addr, RTA_DATA(rta),
-				    sizeof(rt6.dest.s6_addr));
-				break;
-			case RTA_GATEWAY:
-				memcpy(&rt6.gate.s6_addr, RTA_DATA(rta),
-				    sizeof(rt6.gate.s6_addr));
-				break;
-			case RTA_OIF:
-				rt6.iface = if_findindex(ctx,
-				    *(unsigned int *)RTA_DATA(rta));
-				break;
-			}
-			break;
-#endif
-		}
-		switch (rta->rta_type) {
-		case RTA_PRIORITY:
-			metric = *(unsigned int *)RTA_DATA(rta);
-			break;
-		}
-		rta = RTA_NEXT(rta, len);
-	}
-
+	rtm = NLMSG_DATA(nlm);
 	switch (rtm->rtm_family) {
 #ifdef INET
-		case AF_INET:
-			if (rt.iface != NULL && metric == rt.iface->metric) {
-				inet_cidrtoaddr(rtm->rtm_dst_len, &rt.net);
-				ipv4_routedeleted(ctx, &rt);
-			}
-			break;
+	case AF_INET:
+		if (if_copyrt(ctx, &rt, nlm) == 0)
+			ipv4_handlert(ctx, cmd, &rt);
+		break;
 #endif
 #ifdef INET6
-		case AF_INET6:
-			if (rt6.iface != NULL && metric == rt6.iface->metric) {
-				ipv6_mask(&rt6.net, rtm->rtm_dst_len);
-				ipv6_routedeleted(ctx, &rt6);
-			}
-			break;
+	case AF_INET6:
+		if (if_copyrt6(ctx, &rt6, nlm) == 0)
+			ipv6_handlert(ctx, cmd, &rt6);
+		break;
 #endif
 	}
-	return 1;
+
+	return 0;
 }
 
 static int
@@ -541,7 +594,7 @@ link_addr(struct dhcpcd_ctx *ctx, struct interface *ifp, struct nlmsghdr *nlm)
 		 * so it's not really an error */
 		return 1;
 	}
-	rta = (struct rtattr *) IFA_RTA(ifa);
+	rta = (struct rtattr *)IFA_RTA(ifa);
 	len = NLMSG_PAYLOAD(nlm, sizeof(*ifa));
 	switch (ifa->ifa_family) {
 #ifdef INET
@@ -585,7 +638,7 @@ link_addr(struct dhcpcd_ctx *ctx, struct interface *ifp, struct nlmsghdr *nlm)
 		break;
 #endif
 	}
-	return 1;
+	return 0;
 }
 
 static uint8_t
@@ -665,7 +718,7 @@ link_neigh(struct dhcpcd_ctx *ctx, __unused struct interface *ifp,
 		ipv6nd_neighbour(ctx, &addr6, flags);
 	}
 
-	return 1;
+	return 0;
 }
 #endif
 
@@ -700,7 +753,7 @@ link_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp,
 	}
 	ifi = NLMSG_DATA(nlm);
 	if (ifi->ifi_flags & IFF_LOOPBACK)
-		return 1;
+		return 0;
 	rta = (struct rtattr *)(void *)((char *)ifi +NLMSG_ALIGN(sizeof(*ifi)));
 	len = NLMSG_PAYLOAD(nlm, sizeof(*ifi));
 	*ifn = '\0';
@@ -712,7 +765,7 @@ link_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp,
 			/* Ignore wireless messages */
 			if (nlm->nlmsg_type == RTM_NEWLINK &&
 			    ifi->ifi_change == 0)
-				return 1;
+				return 0;
 			break;
 		case IFLA_IFNAME:
 			strlcpy(ifn, RTA_DATA(rta), sizeof(ifn));
@@ -726,7 +779,7 @@ link_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp,
 
 	if (nlm->nlmsg_type == RTM_DELLINK) {
 		dhcpcd_handleinterface(ctx, -1, ifn);
-		return 1;
+		return 0;
 	}
 
 	/* Virtual interfaces may not get a valid hardware address
@@ -735,12 +788,12 @@ link_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp,
 	 * that that don't exist until they have one. */
 	if (ifi->ifi_flags & IFF_MASTER && !hwaddr) {
 		dhcpcd_handleinterface(ctx, -1, ifn);
-		return 1;
+		return 0;
 	}
 
 	/* Check for interface name change */
 	if (handle_rename(ctx, (unsigned int)ifi->ifi_index, ifn))
-		    return 1;
+		return 0;
 
 	/* Check for a new interface */
 	ifp = if_find(ctx, ifn);
@@ -749,7 +802,7 @@ link_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp,
 		 * the interface rather than the kernel. */
 		if (dev_listening(ctx) < 1)
 			dhcpcd_handleinterface(ctx, 1, ifn);
-		return 1;
+		return 0;
 	}
 
 	/* Re-read hardware address and friends */
@@ -764,7 +817,7 @@ link_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp,
 	dhcpcd_handlecarrier(ctx,
 	    ifi->ifi_flags & IFF_RUNNING ? LINK_UP : LINK_DOWN,
 	    ifi->ifi_flags, ifn);
-	return 1;
+	return 0;
 }
 
 int
@@ -778,7 +831,7 @@ if_managelink(struct dhcpcd_ctx *ctx)
 static int
 send_netlink(struct dhcpcd_ctx *ctx, struct interface *ifp,
     int protocol, struct nlmsghdr *hdr,
-    int (*callback)(struct dhcpcd_ctx *, struct interface *,struct nlmsghdr *))
+    int (*callback)(struct dhcpcd_ctx *, struct interface *, struct nlmsghdr *))
 {
 	int s, r;
 	struct sockaddr_nl snl;
@@ -861,7 +914,7 @@ nla_next(struct nlattr *nla, size_t *rem)
 {
 
 	*rem -= NLA_ALIGN(nla->nla_len);
-	return (struct nlattr *)((char *)nla + NLA_ALIGN(nla->nla_len));
+	return (struct nlattr *)(void *)((char *)nla + NLA_ALIGN(nla->nla_len));
 }
 
 #define NLA_TYPE(nla) ((nla)->nla_type & NLA_TYPE_MASK)
@@ -935,7 +988,7 @@ gnl_parse(struct nlmsghdr *nlm, struct nlattr *tb[], int maxtype)
 
 	memset(tb, 0, sizeof(*tb) * ((unsigned int)maxtype + 1));
 	ghdr = NLMSG_DATA(nlm);
-	head = (struct nlattr *)((char *) ghdr + GENL_HDRLEN);
+	head = (struct nlattr *)(void *)((char *) ghdr + GENL_HDRLEN);
 	len = nlm->nlmsg_len - GENL_HDRLEN - NLMSG_HDRLEN;
 	NLA_FOR_EACH_ATTR(nla, head, len, rem) {
 		type = NLA_TYPE(nla);
@@ -959,7 +1012,7 @@ _gnl_getfamily(__unused struct dhcpcd_ctx *ctx, __unused struct interface *ifp,
 		errno = ENOENT;
 		return -1;
 	}
-	family = *(uint16_t *)NLA_DATA(tb[CTRL_ATTR_FAMILY_ID]);
+	family = *(uint16_t *)(void *)NLA_DATA(tb[CTRL_ATTR_FAMILY_ID]);
 	return (int)family;
 }
 
@@ -1252,7 +1305,7 @@ if_address(const struct interface *iface,
 }
 
 int
-if_route(const struct rt *rt, int action)
+if_route(unsigned char cmd, const struct rt *rt)
 {
 	struct nlmr nlm;
 	int retval = 0;
@@ -1260,22 +1313,27 @@ if_route(const struct rt *rt, int action)
 
 	memset(&nlm, 0, sizeof(nlm));
 	nlm.hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
-	nlm.hdr.nlmsg_type = RTM_NEWROUTE;
-	if (action == 0)
-		nlm.hdr.nlmsg_flags = NLM_F_REPLACE;
-	else if (action == 1)
+	switch (cmd) {
+	case RTM_CHANGE:
+		nlm.hdr.nlmsg_type = RTM_NEWROUTE;
+		nlm.hdr.nlmsg_flags = NLM_F_CREATE | NLM_F_REPLACE;
+		break;
+	case RTM_ADD:
+		nlm.hdr.nlmsg_type = RTM_NEWROUTE;
 		nlm.hdr.nlmsg_flags = NLM_F_CREATE | NLM_F_EXCL;
-	else
+		break;
+	case RTM_DELETE:
 		nlm.hdr.nlmsg_type = RTM_DELROUTE;
+		break;
+	}
 	nlm.hdr.nlmsg_flags |= NLM_F_REQUEST;
 	nlm.rt.rtm_family = AF_INET;
 	nlm.rt.rtm_table = RT_TABLE_MAIN;
 
 	state = D_STATE(rt->iface);
-	if (action == -1 || action == -2)
+	if (cmd == RTM_DELETE)
 		nlm.rt.rtm_scope = RT_SCOPE_NOWHERE;
 	else {
-		nlm.hdr.nlmsg_flags |= NLM_F_CREATE | NLM_F_EXCL;
 		/* We only change route metrics for kernel routes */
 		if (rt->dest.s_addr ==
 		    (state->addr.s_addr & state->net.s_addr) &&
@@ -1302,20 +1360,53 @@ if_route(const struct rt *rt, int action)
 		    &state->addr.s_addr, sizeof(state->addr.s_addr));
 	}
 	/* If destination == gateway then don't add the gateway */
-	if (rt->dest.s_addr != rt->gate.s_addr ||
-	    rt->net.s_addr != INADDR_BROADCAST)
+	if ((cmd == RTM_ADD || cmd == RTM_CHANGE) &&
+	    (rt->dest.s_addr != rt->gate.s_addr ||
+	    rt->net.s_addr != INADDR_BROADCAST))
 		add_attr_l(&nlm.hdr, sizeof(nlm), RTA_GATEWAY,
 		    &rt->gate.s_addr, sizeof(rt->gate.s_addr));
 
 	if (rt->gate.s_addr != htonl(INADDR_LOOPBACK))
 		add_attr_32(&nlm.hdr, sizeof(nlm), RTA_OIF, rt->iface->index);
-	add_attr_32(&nlm.hdr, sizeof(nlm), RTA_PRIORITY, rt->metric);
+	if (rt->metric)
+		add_attr_32(&nlm.hdr, sizeof(nlm), RTA_PRIORITY, rt->metric);
 
 	if (send_netlink(rt->iface->ctx, NULL,
 	    NETLINK_ROUTE, &nlm.hdr, NULL) == -1)
 		retval = -1;
 	return retval;
 }
+
+static int
+_if_initrt(struct dhcpcd_ctx *ctx, __unused struct interface *ifp,
+    struct nlmsghdr *nlm)
+{
+	struct rt rt;
+
+	if (if_copyrt(ctx, &rt, nlm) == 0)
+		ipv4_handlert(ctx, RTM_ADD, &rt);
+	return 0;
+}
+
+int
+if_initrt(struct interface *ifp)
+{
+	struct nlmr nlm;
+
+	ipv4_freerts(ifp->ctx->ipv4_kroutes);
+
+	memset(&nlm, 0, sizeof(nlm));
+	nlm.hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
+	nlm.hdr.nlmsg_type = RTM_GETROUTE;
+	nlm.hdr.nlmsg_flags = NLM_F_REQUEST | NLM_F_MATCH;
+	nlm.hdr.nlmsg_flags |= NLM_F_REQUEST;
+	nlm.rt.rtm_family = AF_INET;
+	nlm.rt.rtm_table = RT_TABLE_MAIN;
+	add_attr_32(&nlm.hdr, sizeof(nlm), RTA_OIF, ifp->index);
+
+	return send_netlink(ifp->ctx, ifp,
+	    NETLINK_ROUTE, &nlm.hdr, &_if_initrt);
+}
 #endif
 
 #ifdef INET6
@@ -1327,7 +1418,7 @@ if_address6(const struct ipv6_addr *ap, int action)
 	int retval = 0;
 /* IFA_FLAGS is not a define, but is was added at the same time
  * IFA_F_NOPREFIXROUTE was do use that. */
-#ifdef IFA_F_NOPREFIXROUTE
+#if defined(IFA_F_NOPREFIXROUTE) || defined(IFA_F_MANAGETEMPADDR)
 	uint32_t flags = 0;
 #endif
 
@@ -1374,13 +1465,13 @@ if_address6(const struct ipv6_addr *ap, int action)
 	if (!IN6_IS_ADDR_LINKLOCAL(&ap->addr))
 		flags |= IFA_F_NOPREFIXROUTE;
 #endif
-#ifdef IFA_F_NOPREFIXROUTE
+#if defined(IFA_F_NOPREFIXROUTE) || defined(IFA_F_MANAGETEMPADDR)
 	if (flags)
 		add_attr_32(&nlm.hdr, sizeof(nlm), IFA_FLAGS, flags);
 #endif
 
-	if (send_netlink(ap->iface->ctx, NULL,
-	    NETLINK_ROUTE, &nlm.hdr, NULL) == -1)
+	if (send_netlink(ap->iface->ctx, NULL, NETLINK_ROUTE, &nlm.hdr,
+	    NULL) == -1)
 		retval = -1;
 	return retval;
 }
@@ -1397,7 +1488,8 @@ rta_add_attr_32(struct rtattr *rta, unsigned short maxlen,
 		return -1;
 	}
 
-	subrta = (struct rtattr*)(((char*)rta) + RTA_ALIGN(rta->rta_len));
+	subrta = (struct rtattr*)(void *)
+	    (((char*)rta) + RTA_ALIGN(rta->rta_len));
 	subrta->rta_type = type;
 	subrta->rta_len = len;
 	memcpy(RTA_DATA(subrta), &data, sizeof(data));
@@ -1406,30 +1498,33 @@ rta_add_attr_32(struct rtattr *rta, unsigned short maxlen,
 }
 
 int
-if_route6(const struct rt6 *rt, int action)
+if_route6(unsigned char cmd, const struct rt6 *rt)
 {
 	struct nlmr nlm;
-	char metricsbuf[32];
-	struct rtattr *metrics = (void *)metricsbuf;
 	int retval = 0;
 
 	memset(&nlm, 0, sizeof(nlm));
 	nlm.hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
-	nlm.hdr.nlmsg_type = RTM_NEWROUTE;
-	nlm.hdr.nlmsg_flags = NLM_F_REQUEST;
-	if (action == 0)
-		nlm.hdr.nlmsg_flags |= NLM_F_REPLACE;
-	else if (action == 1)
-		nlm.hdr.nlmsg_flags |= NLM_F_CREATE | NLM_F_EXCL;
-	else
+	switch (cmd) {
+	case RTM_CHANGE:
+		nlm.hdr.nlmsg_type = RTM_NEWROUTE;
+		nlm.hdr.nlmsg_flags = NLM_F_CREATE | NLM_F_REPLACE;
+		break;
+	case RTM_ADD:
+		nlm.hdr.nlmsg_type = RTM_NEWROUTE;
+		nlm.hdr.nlmsg_flags = NLM_F_CREATE | NLM_F_EXCL;
+		break;
+	case RTM_DELETE:
 		nlm.hdr.nlmsg_type = RTM_DELROUTE;
+		break;
+	}
+	nlm.hdr.nlmsg_flags |= NLM_F_REQUEST;
 	nlm.rt.rtm_family = AF_INET6;
 	nlm.rt.rtm_table = RT_TABLE_MAIN;
 
-	if (action == -1 || action == -2)
+	if (cmd == RTM_DELETE)
 		nlm.rt.rtm_scope = RT_SCOPE_NOWHERE;
 	else {
-		nlm.hdr.nlmsg_flags |= NLM_F_CREATE | NLM_F_EXCL;
 		/* None interface subnet routes are static. */
 		if (rt->iface->flags & IFF_LOOPBACK)
 			nlm.rt.rtm_scope = RT_SCOPE_HOST;
@@ -1448,21 +1543,25 @@ if_route6(const struct rt6 *rt, int action)
 	add_attr_l(&nlm.hdr, sizeof(nlm), RTA_DST,
 	    &rt->dest.s6_addr, sizeof(rt->dest.s6_addr));
 
-	if (action >= 0 && !IN6_IS_ADDR_UNSPECIFIED(&rt->gate))
+	if (cmd == RTM_ADD && !IN6_IS_ADDR_UNSPECIFIED(&rt->gate))
 		add_attr_l(&nlm.hdr, sizeof(nlm), RTA_GATEWAY,
 		    &rt->gate.s6_addr, sizeof(rt->gate.s6_addr));
 
 	if (!(rt->flags & RTF_REJECT)) {
 		add_attr_32(&nlm.hdr, sizeof(nlm), RTA_OIF, rt->iface->index);
-		add_attr_32(&nlm.hdr, sizeof(nlm), RTA_PRIORITY, rt->metric);
+		if (rt->metric)
+			add_attr_32(&nlm.hdr, sizeof(nlm),
+			    RTA_PRIORITY, rt->metric);
 	}
+	if (cmd == RTM_ADD && rt->mtu) {
+		char metricsbuf[32];
+		struct rtattr *metrics = (void *)metricsbuf;
 
-	if (rt->mtu) {
 		metrics->rta_type = RTA_METRICS;
 		metrics->rta_len = RTA_LENGTH(0);
 		rta_add_attr_32(metrics, sizeof(metricsbuf), RTAX_MTU, rt->mtu);
 		add_attr_l(&nlm.hdr, sizeof(nlm), RTA_METRICS,
-		    RTA_DATA(metrics), RTA_PAYLOAD(metrics));
+		    RTA_DATA(metrics), (unsigned short)RTA_PAYLOAD(metrics));
 	}
 
 	if (send_netlink(rt->iface->ctx, NULL,
@@ -1471,6 +1570,37 @@ if_route6(const struct rt6 *rt, int action)
 	return retval;
 }
 
+static int
+_if_initrt6(struct dhcpcd_ctx *ctx, __unused struct interface *ifp,
+    struct nlmsghdr *nlm)
+{
+	struct rt6 rt;
+
+	if (if_copyrt6(ctx, &rt, nlm) == 0)
+		ipv6_handlert(ctx, RTM_ADD, &rt);
+	return 0;
+}
+
+int
+if_initrt6(struct interface *ifp)
+{
+	struct nlmr nlm;
+
+	ipv6_freerts(&ifp->ctx->ipv6->kroutes);
+
+	memset(&nlm, 0, sizeof(nlm));
+	nlm.hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct rtmsg));
+	nlm.hdr.nlmsg_type = RTM_GETROUTE;
+	nlm.hdr.nlmsg_flags = NLM_F_REQUEST | NLM_F_MATCH;
+	nlm.hdr.nlmsg_flags |= NLM_F_REQUEST;
+	nlm.rt.rtm_family = AF_INET6;
+	nlm.rt.rtm_table = RT_TABLE_MAIN;
+	add_attr_32(&nlm.hdr, sizeof(nlm), RTA_OIF, ifp->index);
+
+	return send_netlink(ifp->ctx, ifp,
+	    NETLINK_ROUTE, &nlm.hdr, &_if_initrt6);
+}
+
 int
 if_addrflags6(const struct in6_addr *addr, const struct interface *ifp)
 {
@@ -1586,8 +1716,8 @@ if_checkipv6(struct dhcpcd_ctx *ctx, const struct interface *ifp, int own)
 		ifname = "all";
 	else if (own) {
 		if (if_disable_autolinklocal(ctx, (int)ifp->index) == -1)
-			syslog(LOG_DEBUG, "%s: if_disable_autolinklocal: %m",
-			    ifp->name);
+			logger(ctx, LOG_DEBUG,
+			    "%s: if_disable_autolinklocal: %m", ifp->name);
 	}
 	if (ifp)
 		ifname = ifp->name;
@@ -1596,11 +1726,11 @@ if_checkipv6(struct dhcpcd_ctx *ctx, const struct interface *ifp, int own)
 	ra = check_proc_int(path);
 	if (ra != 1) {
 		if (!own)
-			syslog(LOG_WARNING,
+			logger(ctx, LOG_WARNING,
 			    "%s: IPv6 kernel autoconf disabled", ifname);
 	} else if (ra != -1 && own) {
 		if (write_path(path, "0") == -1) {
-			syslog(LOG_ERR, "write_path: %s: %m", path);
+			logger(ctx, LOG_ERR, "write_path: %s: %m", path);
 			return -1;
 		}
 	}
@@ -1610,13 +1740,13 @@ if_checkipv6(struct dhcpcd_ctx *ctx, const struct interface *ifp, int own)
 	if (ra == -1)
 		/* The sysctl probably doesn't exist, but this isn't an
 		 * error as such so just log it and continue */
-		syslog(errno == ENOENT ? LOG_DEBUG : LOG_WARNING,
+		logger(ctx, errno == ENOENT ? LOG_DEBUG : LOG_WARNING,
 		    "%s: %m", path);
 	else if (ra != 0 && own) {
-		syslog(LOG_DEBUG, "%s: disabling kernel IPv6 RA support",
+		logger(ctx, LOG_DEBUG, "%s: disabling kernel IPv6 RA support",
 		    ifname);
 		if (write_path(path, "0") == -1) {
-			syslog(LOG_ERR, "write_path: %s: %m", path);
+			logger(ctx, LOG_ERR, "write_path: %s: %m", path);
 			return ra;
 		}
 		return 0;
@@ -1625,6 +1755,7 @@ if_checkipv6(struct dhcpcd_ctx *ctx, const struct interface *ifp, int own)
 	return ra;
 }
 
+#ifdef IPV6_MANAGETEMPADDR
 int
 ip6_use_tempaddr(const char *ifname)
 {
@@ -1664,4 +1795,5 @@ ip6_temp_valid_lifetime(const char *ifname)
 	val = check_proc_int(path);
 	return val < 0 ? TEMP_VALID_LIFETIME : val;
 }
-#endif
+#endif /* IPV6_MANAGETEMPADDR */
+#endif /* INET6 */
